@@ -419,7 +419,6 @@ def test_class_method_with_metaclass_spy(mocker: MockerFixture) -> None:
         pass
 
     class Foo:
-
         __metaclass__ = MetaFoo
 
         @classmethod
@@ -631,6 +630,86 @@ def test_assert_has_calls(mocker: MockerFixture) -> None:
         stub.assert_has_calls([mocker.call("bar")])
 
 
+def test_assert_has_calls_multiple_calls(mocker: MockerFixture) -> None:
+    stub = mocker.stub()
+    stub("foo")
+    stub("bar")
+    stub("baz")
+    stub.assert_has_calls([mocker.call("foo"), mocker.call("bar"), mocker.call("baz")])
+    with assert_traceback():
+        stub.assert_has_calls(
+            [
+                mocker.call("foo"),
+                mocker.call("bar"),
+                mocker.call("baz"),
+                mocker.call("bat"),
+            ]
+        )
+    with assert_traceback():
+        stub.assert_has_calls(
+            [mocker.call("foo"), mocker.call("baz"), mocker.call("bar")]
+        )
+
+
+def test_assert_has_calls_multiple_calls_subset(mocker: MockerFixture) -> None:
+    stub = mocker.stub()
+    stub("foo")
+    stub("bar")
+    stub("baz")
+    stub.assert_has_calls([mocker.call("bar"), mocker.call("baz")])
+    with assert_traceback():
+        stub.assert_has_calls([mocker.call("foo"), mocker.call("baz")])
+    with assert_traceback():
+        stub.assert_has_calls(
+            [mocker.call("foo"), mocker.call("bar"), mocker.call("bat")]
+        )
+    with assert_traceback():
+        stub.assert_has_calls([mocker.call("baz"), mocker.call("bar")])
+
+
+def test_assert_has_calls_multiple_calls_any_order(mocker: MockerFixture) -> None:
+    stub = mocker.stub()
+    stub("foo")
+    stub("bar")
+    stub("baz")
+    stub.assert_has_calls(
+        [mocker.call("foo"), mocker.call("baz"), mocker.call("bar")], any_order=True
+    )
+    with assert_traceback():
+        stub.assert_has_calls(
+            [
+                mocker.call("foo"),
+                mocker.call("baz"),
+                mocker.call("bar"),
+                mocker.call("bat"),
+            ],
+            any_order=True,
+        )
+
+
+def test_assert_has_calls_multiple_calls_any_order_subset(
+    mocker: MockerFixture,
+) -> None:
+    stub = mocker.stub()
+    stub("foo")
+    stub("bar")
+    stub("baz")
+    stub.assert_has_calls([mocker.call("baz"), mocker.call("foo")], any_order=True)
+    with assert_traceback():
+        stub.assert_has_calls(
+            [mocker.call("baz"), mocker.call("foo"), mocker.call("bat")], any_order=True
+        )
+
+
+def test_assert_has_calls_no_calls(
+    mocker: MockerFixture,
+) -> None:
+    stub = mocker.stub()
+    stub.assert_has_calls([])
+    with assert_traceback():
+        stub.assert_has_calls([mocker.call("foo")])
+
+
 def test_monkeypatch_ini(testdir: Any, mocker: MockerFixture) -> None:
     # Make sure the following function actually tests something
     stub = mocker.stub()
@@ -638,7 +717,6 @@ def test_monkeypatch_ini(testdir: Any, mocker: MockerFixture) -> None:
 
     testdir.makepyfile(
         """
-        import py.code
         def test_foo(mocker):
             stub = mocker.stub()
             assert stub.assert_called_with.__module__ == stub.__module__
@@ -1100,3 +1178,56 @@ def test_used_with_session_scope(testdir: Any) -> None:
     result = testdir.runpytest_subprocess()
     assert "AssertionError" not in result.stderr.str()
     result.stdout.fnmatch_lines("* 1 passed in *")
+
+
+def test_stop_patch(mocker):
+    class UnSpy:
+        def foo(self):
+            return 42
+
+    m = mocker.patch.object(UnSpy, "foo", return_value=0)
+    assert UnSpy().foo() == 0
+    mocker.stop(m)
+    assert UnSpy().foo() == 42
+
+    with pytest.raises(ValueError):
+        mocker.stop(m)
+
+
+def test_stop_instance_patch(mocker):
+    class UnSpy:
+        def foo(self):
+            return 42
+
+    m = mocker.patch.object(UnSpy, "foo", return_value=0)
+    un_spy = UnSpy()
+    assert un_spy.foo() == 0
+    mocker.stop(m)
+    assert un_spy.foo() == 42
+
+
+def test_stop_spy(mocker):
+    class UnSpy:
+        def foo(self):
+            return 42
+
+    spy = mocker.spy(UnSpy, "foo")
+    assert UnSpy().foo() == 42
+    assert spy.call_count == 1
+    mocker.stop(spy)
+    assert UnSpy().foo() == 42
+    assert spy.call_count == 1
+
+
+def test_stop_instance_spy(mocker):
+    class UnSpy:
+        def foo(self):
+            return 42
+
+    spy = mocker.spy(UnSpy, "foo")
+    un_spy = UnSpy()
+    assert un_spy.foo() == 42
+    assert spy.call_count == 1
+    mocker.stop(spy)
+    assert un_spy.foo() == 42
+    assert spy.call_count == 1
